﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Controls;

namespace FrequencyConverterMonitor
{
    public class DeviceInfo
    {
        public string Name { get; set; }
        public int DeviceNumber { get; set; }
        public DeviceConfig Config { get; set; }
        public bool Active { get; set; }
        public bool IsShow { get; set; }
        public bool IsShowSetting { get; set; }
        public ComminicationState State { get; set; }

        public FCConfigControl ConfigControl { get; set; }
        public FCSlide SlideControl { get; set; }
        public Button ShowButton { get; set; }
        public TabItem SettingTab { get; set; }
        public MainWindow MainWindow;

        List<SlideParam> Params = new List<SlideParam>();

        public void Init(string name, int deviceNumber, bool active, DeviceConfig config)
        {
            Name = name;
            DeviceNumber = deviceNumber;
            Active = active;
            Config = config;

            Params.Clear();
            foreach (SlideParam p in Config.MainParams) Params.Add(p);
            foreach (SlideParam p in Config.SlideParams) Params.Add(p);
            foreach (SlideParam p in Config.Inputs) Params.Add(p);
            foreach (SlideParam p in Config.Outputs) Params.Add(p);

            if (SettingTab != null) SettingTab.Header = Name;
            if (ConfigControl != null) ConfigControl.Init(this);
            if (ShowButton != null) ShowButton.Content = Name;
            if (SlideControl != null) SlideControl.Init(this);
        }

        public void RunCmd(CommandType cmd)
        {
            MainWindow.RunCmd(this, cmd);
        }

        public bool RunCmd(SerialPortContener port, CommandType cmd)
        {
            var c = Config.Commands.Find(t => t.Type == cmd);
            if (c == null) return false;
            return WriteRequest(port, (int)c.ModbusFunction, (int)c.RegNumber, (ushort)c.Value) == ComminicationState.Normal;
        }

        int t = 0;
        public string NextText()
        {
            t = (t + 1) % 100000;
            return t.ToString();
        }

        public void OnTimer()
        {
            try
            {
                SlideControl?.UpdateState(Active, State);

                foreach (var p in Params)
                {
                    lock (p)
                    {
                        if (Active)
                        {
                            p.UpdateControl(State);
                        }
                        else
                        {
                            p.UpdateControl(ComminicationState.None);
                        }
                    }
                }
            }
            catch { }
        }

        public void Hide()
        {
            IsShow = false;
            if (ShowButton == null)
            {
                ShowButton = new Button();
                ShowButton.Margin = new System.Windows.Thickness(3);
                ShowButton.Content = this.Name;
                ShowButton.Click += ShowButton_Click;

                MainWindow.SleepPanel.Children.Add(ShowButton);
            }
            else
            {
                if (ShowButton.Parent == null)
                {
                    MainWindow.SleepPanel.Children.Add(ShowButton);
                }
            }

            if (SlideControl != null && SlideControl.Parent != null) MainWindow.MainSlidePanel.Children.Remove(SlideControl);
        }

        public void Close()
        {
            MainWindow.DeleteDevice(this);
        }

        public void Show()
        {
            if (SlideControl == null)
            {
                SlideControl = new FCSlide();
                SlideControl.Init(this);
            }

            IsShow = true;

            if (SlideControl.Parent == null)
            {
                MainWindow.MainSlidePanel.Children.Add(SlideControl);
            }

            if (ShowButton != null && ShowButton.Parent != null)
            {
                MainWindow.SleepPanel.Children.Remove(ShowButton);
            }
        }

        public void ShowSettings()
        {
            if (SettingTab == null)
            {
                SettingTab = new TabItem();
                SettingTab.Header = Name;

                ConfigControl = new FCConfigControl();
                ConfigControl.Init(this);
                SettingTab.Content = ConfigControl;
            }

            if (SettingTab.Parent == null)
            {
                MainWindow.MainTabControl.Items.Add(SettingTab);
                ConfigControl.ReadDeviceConfigs();
            }

            MainWindow.MainTabControl.SelectedItem = SettingTab;
            IsShowSetting = true;
        }

        internal void CloseSettingsTab()
        {
            MainWindow.MainTabControl.Items.Remove(SettingTab);
            SettingTab = null;
            ConfigControl = null;
            IsShowSetting = false;
        }

        private void ShowButton_Click(object sender, System.Windows.RoutedEventArgs e)
        {
            Show();
        }

        Random _random = new Random(System.Environment.TickCount);
        public void SimUpdate()
        {
            try
            {
                State = ComminicationState.Normal;

                foreach (RequestInfo r in this.Config.Reqeusts)
                {
                    for (int i = 0; i < r.RegCount; i++)
                    {
                        int address = (int)r.RegNumber + i;
                        ushort data = (ushort)_random.Next(0, 1000);

                        if (address == 0x3000) data = (ushort)(1 + data % 3);
                        else if (address == 0x8000) data = (ushort)(data % 30);

                        foreach (var p in Params) p.UpdateValue(r.ModbusFunction, address, data);
                        foreach (var p in Config.Errors) p.UpdateValue(r.ModbusFunction, address, data);
                    }
                }
            }
            catch { }
        }

        public int Update(List<ParamInfo> plist, int index)
        {
            try { return MainWindow.Update(this, plist, index); }
            catch { return 0; }
        }

        public int WriteParam(ParamInfo pInfo)
        {
            try { return MainWindow.WriteParam(this, pInfo); }
            catch { return 0; }
        }

        public bool WriteParamValue(int f, int reg, ushort data)
        {
            try { return MainWindow.WriteParam(this, f, reg, data); }
            catch { return false; }
        }

        public bool WriteParam(ManualResetEvent exitEvent, SerialPortContener comPort, int ModbusWriteFunction, int RegNumber, ushort data)
        {
            if (exitEvent.WaitOne(1)) return false;
            if (WriteRequest(comPort, ModbusWriteFunction, RegNumber, data) == ComminicationState.Normal)
            {
                return true;
            }
            return false;
        }

        public int WriteParam(System.Threading.ManualResetEvent exitEvent, SerialPortContener port, ParamInfo pInfo)
        {
            if (exitEvent.WaitOne(1)) return 0;
            ushort data = pInfo.ModbusValueU4 ?? 0;
            if (WriteRequest(port, pInfo.ModbusWriteFunction, pInfo.WriteRegNumber, data) == ComminicationState.Normal)
            {
                return 1;
            }
            return 0;
        }

        public int Update(System.Threading.ManualResetEvent exitEvent, SerialPortContener port, List<ParamInfo> plist, int index)
        {
            if (exitEvent.WaitOne(1)) return 0;

            ushort[] data = null;
            if (ReadRequest(port, plist[index].ModbusReadFunction, plist[index].ReadRegNumber, plist[index].RegCount, out data) == ComminicationState.Normal)
            {
                plist[index].SetValue(data, 0);
                return 1;
            }
            return 0;
        }

        public void Update(System.Threading.ManualResetEvent exitEvent, SerialPortContener port)
        {
            try
            {
                foreach (RequestInfo r in this.Config.Reqeusts)
                {
                    if (exitEvent.WaitOne(1)) return;

                    lock (port)
                    {
                        State = ReadRequest(port, r);
                    }
                }
            }
            catch
            {
                State = ComminicationState.PortError;
            }
        }

        static ushort CRC16(byte[] buffer, int index, int length)
        {
            ushort crc = 0xFFFF;
            char crcLsb;
            for (int i = 0; i < length; i++)
            {
                crc = (ushort)(crc ^ buffer[index + i]);
                for (int j = 0; j < 8; j++)
                {
                    crcLsb = (char)(crc & 0x0001);
                    crc = (ushort)((crc >> 1) & 0x7FFF);
                    if (crcLsb == 1) crc = (ushort)(crc ^ 0xA001);
                }
            }
            return crc;
        }

       

        public ComminicationState WriteRequest(SerialPortContener port, int modbusFunction, int regNumber, ushort data)
        {
            lock (port)
            {
                if (modbusFunction == 6) //Preset Single Register
                {
                    #region F6 Preset Single Register

                    int request_size = 8;
                    byte[] request = new byte[request_size];

                    request[0] = (byte)DeviceNumber;
                    request[1] = (byte)modbusFunction;
                    request[2] = (byte)(regNumber >> 8);
                    request[3] = (byte)(regNumber & 0xFF);
                    request[4] = (byte)(data >> 8);
                    request[5] = (byte)(data & 0xFF);

                    ushort crc = CRC16(request, 0, request_size - 2);
                    request[request_size - 2] = (byte)(crc & 0xff);
                    request[request_size - 1] = (byte)((crc >> 8) & 0xff);

                    port.DiscardInBuffer();
                    port.Write(request, 0, request.Length);

                    MainWindow.WriteToLog(false, request, request.Length);

                    int response_size = request_size;
                    byte[] response = new byte[response_size];
                    Stopwatch timer = Stopwatch.StartNew();
                    int readCount = 0;
                    do
                    {
                        try
                        {
                            int t = port.Read(response, readCount, response.Length - readCount);

                            if (t < 0 && timer.ElapsedMilliseconds > 2000)
                            {
                                if (readCount > 0)
                                {
                                    MainWindow.WriteToLog(true, response, readCount);
                                    return ComminicationState.ErrorResponse;
                                }
                                return ComminicationState.NotResponse;
                            }

                            if (t > 0) readCount += t;

                            if (readCount == 5 && response[0] == request[0] && response[1] == 131)
                            {
                                ushort crc_response_error = CRC16(response, 0, readCount - 2);

                                if (response[readCount - 2] == (byte)(crc_response_error & 0xff) &&
                                    response[readCount - 1] == (byte)((crc_response_error >> 8) & 0xff))
                                {
                                    MainWindow.WriteToLog(true, response, readCount);
                                    return ComminicationState.ErrorResponse;
                                }
                            }
                        }
                        catch
                        {
                            if (readCount > 0)
                            {
                                MainWindow.WriteToLog(true, response, readCount);
                                return ComminicationState.ErrorResponse;
                            }
                            return ComminicationState.NotResponse;
                        }
                    }
                    while (readCount < response_size);

                    MainWindow.WriteToLog(true, response, response_size);

                    ushort crc_response = CRC16(response, 0, response_size - 2);

                    if (response[0] == request[0] &&
                        response[1] == request[1] &&
                        response[2] == request[2] &&
                        response[3] == request[3] &&
                        response[response_size - 2] == (byte)(crc_response & 0xff) &&
                        response[response_size - 1] == (byte)((crc_response >> 8) & 0xff))
                    {
                        return ComminicationState.Normal;
                    }
                    else
                    {
                        return ComminicationState.NotResponse;
                    }
                    #endregion
                }
                else
                {
                    return ComminicationState.None;
                }
            }
        }

        protected ComminicationState ReadRequest(SerialPortContener port, long ModbusFunction, long RegNumber, long RegCount, out ushort[] data)
        {
            lock (port)
            {
                if (ModbusFunction == 3) //Read Holding Registers
                {
                    #region F3 Read Holding Registers

                    int request_size = 8;
                    byte[] request = new byte[request_size];
                    int start = (int)RegNumber;
                    int count = (int)RegCount;

                    request[0] = (byte)DeviceNumber;
                    request[1] = 3;
                    request[2] = (byte)(start >> 8);
                    request[3] = (byte)(start & 0xFF);
                    request[4] = (byte)(count >> 8);
                    request[5] = (byte)(count & 0xFF);

                    ushort crc = CRC16(request, 0, request_size - 2);
                    request[request_size - 2] = (byte)(crc & 0xff);
                    request[request_size - 1] = (byte)((crc >> 8) & 0xff);

                    port.DiscardInBuffer();
                    port.Write(request, 0, request.Length);

                    MainWindow.WriteToLog(false, request, request.Length);

                    int response_size = 5 + (int)count * 2;
                    byte[] response = new byte[response_size];
                    Stopwatch timer = Stopwatch.StartNew();
                    int readCount = 0;
                    do
                    {
                        try
                        {
                            int t = port.Read(response, readCount, response.Length - readCount);

                            if (t < 0 && timer.ElapsedMilliseconds > 2000)
                            {
                                if (readCount > 0)
                                {
                                    MainWindow.WriteToLog(true, response, readCount);
                                    data = null;
                                    return ComminicationState.ErrorResponse;
                                }
                                data = null;
                                return ComminicationState.NotResponse;
                            }

                            if (t > 0) readCount += t;

                            if (readCount == 5 && response[0] == request[0] && response[1] == 131)
                            {
                                ushort crc_response_error = CRC16(response, 0, readCount - 2);

                                if (response[readCount - 2] == (byte)(crc_response_error & 0xff) &&
                                    response[readCount - 1] == (byte)((crc_response_error >> 8) & 0xff))
                                {
                                    MainWindow.WriteToLog(true, response, readCount);
                                    data = null;
                                    return ComminicationState.ErrorResponse;
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            data = null;
                            if (readCount > 0)
                            {
                                MainWindow.WriteToLog(true, response, readCount);
                                MainWindow.WriteToLog(ex);
                                return ComminicationState.ErrorResponse;
                            }
                            MainWindow.WriteToLog(ex);
                            return ComminicationState.NotResponse;
                        }
                    }
                    while (readCount < response_size);

                    MainWindow.WriteToLog(true, response, response_size);

                    ushort crc_response = CRC16(response, 0, response_size - 2);

                    if (response[0] == request[0] &&
                        response[1] == request[1] &&
                        response[2] == count * 2 &&
                        response[response_size - 2] == (byte)(crc_response & 0xff) &&
                        response[response_size - 1] == (byte)((crc_response >> 8) & 0xff))
                    {
                        data = new ushort[count];
                        for (int i = 0; i < count; i++)
                        {
                            int address = start + i;
                            data[i] = (ushort)(((ushort)response[3 + i * 2] << 8) | response[3 + i * 2 + 1]);
                        }
                        return ComminicationState.Normal;
                    }
                    else
                    {
                        data = null;
                        return ComminicationState.NotResponse;
                    }
                    #endregion
                }
                else
                {
                    data = null;
                    return ComminicationState.None;
                }
            }
        }

        protected ComminicationState ReadRequest(SerialPortContener port, RequestInfo r)
        {
            if (r.ModbusFunction == 3) //Read Holding Registers
            {
                #region F3 Read Holding Registers
                ushort[] tmpdata;
                var s = ReadRequest(port, (int)r.ModbusFunction, (int)r.RegNumber, (int)r.RegCount, out tmpdata);
                if (s == ComminicationState.Normal)
                {
                    for (int i = 0; i < r.RegCount; i++)
                    {
                        int address = (int)r.RegNumber + i;
                        ushort data = tmpdata[i];
                        foreach (var p in Params) p.UpdateValue(r.ModbusFunction, address, data);
                        foreach (var p in Config.Errors) p.UpdateValue(r.ModbusFunction, address, data);
                    }
                }
                else
                {
                    for (int i = 0; i < r.RegCount; i++)
                    {
                        int address = (int)r.RegNumber + i;
                        foreach (var p in Params) p.SetErrorValue(r.ModbusFunction, address);
                        foreach (var p in Config.Errors) p.SetErrorValue(r.ModbusFunction, address);
                    }
                }
                return s;
                #endregion
            }
            else
            {
                return ComminicationState.None;
            }
        }
    }

    public enum ComminicationState
    {
        None,
        Fault,
        Normal,
        ErrorResponse,
        NotResponse,
        PortError,
    }
}
