﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Globalization;
using System.ComponentModel;

namespace FrequencyConverterMonitor
{
    /// <summary>
    /// Interactionlogic for NumericUpDown.xaml
    /// </summary>
    public partial class NumericUpDown : UserControl
    {
        double _minValue = 0;
        double _maxValue = 65535.0;

        double? _value = null;
        double _step = 1;

        public IParamInfo Param { get; set; }

        public void LoadFromParamInfo(IParamInfo param)
        {
            Param = param;
            UpdateParamInfo(false);
        }

        public void UpdateParamInfo(bool updateValue)
        {
            this.Unit = Param.Unit;
            if (Param.MinValue != null && Param.MaxValue != null) this.Unit += $" [ {Param.MinValue} ... {Param.MaxValue} ]";
            if (Param.MinValue.HasValue) this.MinValue = Param.MinValue.Value;
            if (Param.MaxValue.HasValue) this.MaxValue = Param.MaxValue.Value;

            if (!string.IsNullOrWhiteSpace(Param.Format))
            {
                this.FormatString = Param.Format;
            }
            else if (!Param.K.HasValue || Param.K == 1)
            {
                this.FormatString = "0";
            }
            else if (Param.K.HasValue && Param.K == 0.1)
            {
                this.FormatString = "0.0";
            }
            else if (Param.K.HasValue && Param.K == 0.01)
            {
                this.FormatString = "0.00";
            }
            else if (Param.K.HasValue && Param.K == 0.001)
            {
                this.FormatString = "0.000";
            }
            this.Step = Param.K ?? 1.0;

            if (updateValue)
            {
                double? v = null;
                if (Param.ModbusValue != null) v = Math.Round(Param.ModbusValue.Value * (Param.K ?? 1.0), 6);
                SetValue(v);
            }
        }

        /// <summary>
        /// Default Constructor (nothing special here x) )
        /// </summary>
        public NumericUpDown()
        {
            InitializeComponent();
        }

        public string TextValue
        {
            get { return tbValue.Text; }
            set { tbValue.Text = value; }
        }

        public string FormatValue(double? v)
        {
            if (!v.HasValue || double.IsNaN(v.Value)) return "";
            return v?.ToString(FormatString) ?? "";
        }

        public double? ParseValue(string text, bool check)
        {
            if (string.IsNullOrWhiteSpace(text)) return null;
            if (double.TryParse(text.Replace(",", "."), NumberStyles.Any, CultureInfo.InvariantCulture, out double v))
            {
                if (check)
                {
                    if (v < _minValue) return _minValue;
                    if (v > _maxValue) return _maxValue;
                }
                return v;
            }
            return null;
        }

        /// <summary>
        /// Gets / Sets the value that the control is showing
        /// </summary>
        /// <exception cref="ArgumentException" />
        /// <remarks>If Value exceeds <see cref="MaxValue"/> or falls below <see cref="MinValue"/>, an <see cref="ArgumentException"/> is thrown</remarks>
        public double? Value
        {
            get
            {
                return _value;
            }
            set
            {
                double? tmp = _value;
                SetValue(value);
                if (Param != null) Param.ModbusValue = ToModbusValue(_value);
                if (tmp != value) ChangedValue?.Invoke(this, null);
            }
        }

        public double? SetValue(double? value)
        {
            _value = value;
            //if (_value.HasValue && _value.Value < _minValue) _value = _minValue;
            //if (_value.HasValue && _value.Value > _maxValue) _value = _maxValue;
            string str = FormatValue(_value);
            double? test = ParseValue(tbValue.Text, false);
            if (_value != test)
            {
                tbValue.Text = str;
            }

            if (_value < _minValue || _value > _maxValue)
            {
                tbValue.Background = Brushes.Red;
            }
            else
            {
                tbValue.Background = Brushes.White;
            }
            return _value;
        }

        public event EventHandler ChangedValue;

        public string Unit
        {
            get { return tbUnit.Text; }
            set { tbUnit.Text = value; }
        }

        /// <summary>
        /// Gets / Sets the minimal value of the control's value
        /// </summary>
        public double MinValue
        {
            get
            {
                return _minValue;
            }
            set
            {
                _minValue = value;
            }
        }

        /// <summary>
        /// Gets / Sets the maximal value of the control's value
        /// </summary>
        public double MaxValue
        {
            get
            {
                return _maxValue;
            }
            set
            {
                _maxValue = value;
            }
        }

        public string FormatString
        {
            get;
            set;
        }

        /// <summary>
        /// Gets / Sets the step size (increment / decrement size) of the control's value
        /// </summary>
        public double Step
        {
            get
            {
                return _step;
            }
            set
            {
                _step = value;
            }
        }


        public bool IsReadOnly
        {
            get { return tbValue.IsReadOnly; }
            set
            {
                tbValue.IsReadOnly = value;
                btnUp.IsEnabled = !value;
                btnDown.IsEnabled = !value;
            }
        }


        /// <summary>
        /// Increments the control's value by the value defined by <see cref="Step"/>
        /// </summary>
        /// <remarks>The value doesn't increment over MaxValue or under MinValue</remarks>
        public void Increment()
        {
            try
            {
                if (!_value.HasValue || double.IsNaN(_value.Value))
                {
                    Value = this.MinValue;
                }
                else
                {
                    Value += Step;
                }
            }
            catch (ArgumentException)
            {
            }
        }

        /// <summary>
        /// Decrements the control's value by the value defined by <see cref="Step"/>
        /// </summary>
        /// <remarks>The value doesn't increment over MaxValue or under MinValue</remarks>
        public void Decrement()
        {
            try
            {
                if (!_value.HasValue || double.IsNaN(_value.Value))
                {
                    Value = this.MaxValue;
                }
                else
                {
                    Value -= Step;
                }
            }
            catch (ArgumentException)
            {
            }
        }

        /// <summary>
        /// Handler for the Up Button Click.
        /// Increments the <see cref="Value"/> by <see cref="Step"/>
        /// </summary>
        /// <param name="sender">The Up Button Control</param>
        /// <param name="e"></param>
        private void btnUp_Click(object sender, RoutedEventArgs e)
        {
            Increment();
        }

        /// <summary>
        /// Handler for the Down Button Click.
        /// Decrements the <see cref="Value"/> by <see cref="Step"/>
        /// </summary>
        /// <param name="sender">The Down Button Control</param>
        /// <param name="e"></param>
        private void btnDown_Click(object sender, RoutedEventArgs e)
        {
            Decrement();
        }

        /// <summary>
        /// Sets the decimation binding.
        /// </summary>
        /// <param name="decimation">The decimation.</param>
        private void SetDecimationBinding(uint decimation)
        {

        }

        /// <summary>
        /// Handles the Loaded event of the ucNumericUpDown control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.RoutedEventArgs"/> instance containing the event data.</param>
        private void ucNumericUpDown_Loaded(object sender, RoutedEventArgs e)
        {

        }

        bool _inTextChange = false;
        private void tbValue_TextChanged(object sender, TextChangedEventArgs e)
        {
            if (_inTextChange) return;
            try
            {
                _inTextChange = true;
                _value = ParseValue(tbValue.Text, true);
                if (Param != null) Param.ModbusValue = ToModbusValue(_value);
            }
            finally
            {
                _inTextChange = false;
            }
        }

        public int? ToModbusValue(double? v)
        {
            if (!v.HasValue) return null;
            if (Param != null) return (int)Math.Round(_value.Value / (Param.K ?? 1));
            return (int)v.Value;
        }

        private void tbValue_LostFocus(object sender, RoutedEventArgs e)
        {
            if (_inTextChange) return;
            try
            {
                _inTextChange = true;
                string str = FormatValue(ParseValue(tbValue.Text, true));
                _value = ParseValue(str, true);
                if (Param != null) Param.ModbusValue = ToModbusValue(_value);
                TextValue = str;
            }
            finally
            {
                _inTextChange = false;
            }
        }
    }
}
